# -*- python -*-
# $RCSfile: meshbdymenu.py,v $
# $Revision: 1.14 $
# $Author: vrc $
# $Date: 2007-09-26 22:03:00 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Menu items for post-solution actions on boundaries.
from ooflib.common.IO import automatic
from ooflib.common.IO import oofmenu
from ooflib.common.IO import parameter
from ooflib.common.IO import destinationparam
from ooflib.common.IO import whoville
from ooflib.engine.IO import meshmenu
from ooflib.engine.IO import meshparameters
from ooflib.engine import meshbdyanalysis
import ooflib.engine.mesh
import types
import string

def _meshBdyAnalyze(menuitem, mesh, boundary, analyzer, destination):
    femesh = ooflib.engine.mesh.meshes[mesh].getObject()
    edgeset = femesh.getBoundary(boundary).edgeset
    outfile = destinationparam.get_file(destination)
    result = analyzer(femesh, edgeset)
    print >> outfile, result
    
analyze = meshmenu.meshmenu.addItem(oofmenu.OOFMenuItem(
    'Boundary_Analysis',
    secret=1,
    callback=_meshBdyAnalyze,
    params = [
    whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                          tip=parameter.emptyTipString),
    meshparameters.MeshEdgeBdyParameter('boundary',
                                        tip="Target boundary."),
    parameter.RegisteredParameter('analyzer', meshbdyanalysis.MeshBdyAnalyzer,
                                  tip="Operation to perform on the boundary."),
    destinationparam.DestinationParameter('destination',
                                          value=automatic.automatic,
                                          tip="Where to put the result?")
    ],
    help="Post-solution analysis of boundaries.",
    discussion="""<para>
    The <command>OOF.Mesh.Boundary_Analysis</command> command runs
    post-solution analyses on the &mesh; boundaries.  Compare to <xref
    linkend='MenuItem:OOF.Mesh.Analyze'/>, which contains commands
    that operate on the interior of the &mesh;.
    </para>"""
    ))




### "mesh" and "boundary" are strings, "flux" is an object, and
### "destination" is a filename or "automatic".
##def _integrateFlux(menuitem, mesh, boundary, flux, destination):
##    fe_mesh = ooflib.engine.mesh.meshes[mesh].getObject()
##    edgeset = fe_mesh.getBoundary(boundary).edgeset
##    outfile = destinationparam.get_file(destination)
##    res = bdyflux.integrate(fe_mesh, flux, edgeset)
##    print >> outfile, res

##meshbdymenu.addItem(oofmenu.OOFMenuItem(
##    'Integrate_Flux',
##    callback=_integrateFlux,
##    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
##                                  tip=parameter.emptyTipString),
##            meshparameters.MeshEdgeBdyParameter('boundary',
##                                                tip="Target boundary."),
##            meshparameters.FluxParameter('flux', tip=parameter.emptyTipString),
##            destinationparam.DestinationParameter('destination',
##                                                  value=automatic.automatic,
##                                                 tip="Where to put the result?")
##            ],
##    help="Integrate a Flux over a specified boundary.",
##    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/integrate_flux.xml')
##    ))
