// -*- C++ -*-
// $RCSfile: celectricfield.C,v $
// $Revision: 1.6 $
// $Author: vrc $
// $Date: 2007-09-26 22:02:35 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "engine/celectricfield.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/csubproblem.h"
#include "engine/field.h"
#include "engine/mastercoord.h"
#include "engine/outputval.h"

void findElectricField(const FEMesh *mesh, const Element *element,
		       const MasterPosition &pos, VECTOR_D &efield)
{
  static ScalarField *voltage =
    dynamic_cast<ScalarField*>(Field::getField("Voltage"));;
  static Field *voltage_z = voltage->out_of_plane();

  efield(0) = 0.0;
  efield(1) = 0.0;
  efield(2) = 0.0;

  bool inplane = voltage->in_plane(mesh);
  for(ElementFuncNodeIterator nu=element->funcnode_iterator(); !nu.end(); ++nu){
    if(nu.hasField(*voltage)) {
      // Shape function derivatives at postion pos.
      double dsfx = nu.dshapefunction(0, pos);
      double dsfy = nu.dshapefunction(1, pos);
      // Displacement components at node nu.
      double phi = (*voltage)(nu, 0)->value();
      efield(0) -= phi*dsfx;
      efield(1) -= phi*dsfy;

      if(!inplane) {
	double sf = nu.shapefunction(pos);
	efield(2) -= sf*(*voltage_z)(nu, 0)->value();
      }
    }
  }
}


