// -*- C++ -*-
// $RCSfile: timederivative.h,v $
// $Revision: 1.7 $
// $Author: vrc $
// $Date: 2007-09-26 22:02:56 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modifed
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef TIMEDERIVATIVE_H
#define TIMEDERIVATIVE_H

#include "common/identification.h"
#include "common/pythonexportable.h"
#include "engine/indextypes.h"
#include "engine/planarity.h"
#include "engine/linearizedsystem.h"
#include "common/space.h"
#include <string>
#include <vector>

class Equation;
class GaussPoint;
class CSubProblem;
class Material;
class NodalEquation;
class FuncNode;

class TimeDerivative {
protected:
  std::vector<std::string> prop_category;//default is "<Unit>"
  /*
    Size of prop_category and fieldList must be the same.
    every entry on prop_category corresponds to the same
    entry number in fieldList. 
    The entries are set at object construction time.
  */
  int order; //default is zero
  Equation *equation; //different from Null when Equation is constructed
  std::vector<std::string> fieldList; //set at object construction time
public:
  TimeDerivative();
  virtual ~TimeDerivative();
  void set_order(int o) {order = o;}
  int get_order() const {return order;}
  //std::string get_prop_category() {return prop_category;}
  void set_equation(Equation * eqn) {equation = eqn;}
  void add_prop_category(const std::string  &prop);
  void add_field(const std::string  &);
  virtual void make_matrix(CSubProblem*, 
			   const Element*, 
			   const GaussPoint&,
			   LinearizedSystem &) const;
  virtual void make_conductivity_matrix(CSubProblem*, 
					const Element*, 
					const GaussPoint&,
					LinearizedSystem &) const;
  //each field contributes its prefactor, 
  //which is defined in a per subclass basis
  const std::vector<Property*> get_properties(const Material &);
  const std::vector<const Field*> get_fields() const;
  virtual double fetch_property(const Field &f, const Material &) const = 0;
  

};


class ScalarKineticScalarTimeDerivative : public TimeDerivative {
public:
  ScalarKineticScalarTimeDerivative(const std::string &, const std::string &);
  virtual ~ScalarKineticScalarTimeDerivative();
  virtual double fetch_property(const Field &, const Material &) const;

};

class ScalarDynamicVectorialTimeDerivative : public TimeDerivative {
public:
  ScalarDynamicVectorialTimeDerivative (const std::string &, const std::string &);
  virtual ~ScalarDynamicVectorialTimeDerivative();
  virtual double fetch_property(const Field &, const Material &) const;
};

#endif
