// -*- C++ -*-
// $RCSfile: timestepadapter.h,v $
// $Revision: 1.11 $
// $Author: vrc $
// $Date: 2007-09-26 22:02:57 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#ifndef TIMESTEPADAPTER_H
#define TIMESTEPADAPTER_H

class SteppingScheme;
class SolverDriverCore;
class CSubProblem;


class TimeStepAdapter {
private:
  double tolerance;
protected:
  double dt; // suggested dt
  bool continew; //flag to stop/continue the computation. default is true
  VECTOR_D * x_prev;/*
		      Solution from the previous step.
		      Will be restored if step adapter fails
		    */
public:
  TimeStepAdapter() : tolerance(0), dt(0), continew(true), x_prev(0) {}
  TimeStepAdapter(const double tol) : tolerance(tol), dt(0), continew(true), x_prev(0) {}
  virtual ~TimeStepAdapter();
  void set_tolerance(const double teef) {tolerance = teef;}
  double get_tolerance() {return tolerance;}
  double get_dt() {return dt;}
  virtual bool is_okay(SteppingScheme *smethod) =0;
  
  
  // hooks for subclasses
  // solution is accepted and time index is updated:
  virtual void accept(CSubProblem*, SteppingScheme*);
  // solution is rejected. time index is NOT updated:
  virtual void reject(CSubProblem*);
  
  virtual void set_time_step(CSubProblem*, SteppingScheme*) = 0;
  // each subclass knows how to set the residual:
  virtual double get_residual(SteppingScheme*) = 0;
  virtual bool keep_going() = 0;
};



class NullAdapter : public TimeStepAdapter {
public:
  NullAdapter() : TimeStepAdapter() {}
  virtual ~NullAdapter() {}
  
  virtual void set_time_step(CSubProblem*, SteppingScheme*);
  virtual double get_residual(SteppingScheme*); // returns zero
  virtual void accept(CSubProblem*, SteppingScheme*);
  virtual void reject(CSubProblem*);
  virtual bool keep_going() { return continew; }
  virtual bool is_okay(SteppingScheme*);
};


#endif // TIMESTEPADAPTER_H
